import * as Rx from 'rxjs';
export declare const State: {
    OK: string;
    ERROR: string;
};
export declare class QEntryRun {
    started: Date;
    _completed: Date;
    _error: any;
    constructor();
    error(err: any): void;
    completed(): void;
}
export declare class QEntry<T> {
    created: Date;
    runs: QEntryRun[];
    input: Rx.Observable<T>;
    output: Rx.Subject<T>;
    constructor(input: Rx.Observable<T>, output: Rx.Subject<T>);
    isCompleted(): boolean;
    isWaiting(): boolean;
    newQEntryRun(): QEntryRun;
}
export interface QConfig {
    taskTimer?: number;
    reclaimTimeout?: number;
    retryWaitTime?: number;
    maxExecuteCnt?: number;
}
export declare class LogMsg {
    readonly level: string;
    readonly parts: any[];
    constructor(level: string, parts: any[]);
}
export interface QWorker<T> {
    (input: Rx.Observable<T>, output: Rx.Subject<T>): void;
}
export declare class WorkerState<T> {
    private running;
    private output;
    private readonly qWorker;
    private readonly q;
    constructor(q: Queue<T>, qw: QWorker<T>);
    isFree(): boolean;
    run(qe: QEntry<any>): void;
}
export declare type QueueSubject<T> = Rx.Subject<Queue<T> | LogMsg>;
export declare type QueueObserver<T> = Rx.Observer<Queue<T> | LogMsg>;
export declare type QueueObservable<T> = Rx.Observable<Queue<T> | LogMsg>;
export declare class Logger<T> {
    private readonly upStream;
    private readonly base;
    constructor(upStream: QueueSubject<T>, qid: number);
    info(...args: any[]): void;
    error(...args: any[]): void;
    debug(...args: any[]): void;
}
export declare class Queue<T> {
    qEntries: QEntry<any>[];
    qTask: number;
    reclaimTimeout: number;
    taskTimer: number;
    retryWaitTime: number;
    maxExecuteCnt: number;
    readonly workers: WorkerState<T>[];
    private readonly lengthSubject;
    readonly logger: Logger<T>;
    constructor(obs: QueueSubject<T>, argv: QConfig, id: number);
    addWorker(qw: QWorker<T>): Queue<T>;
    length(): Rx.Subject<number>;
    stop(): Rx.Observable<void>;
    processMemoryQ(): void;
    push<Y>(action: Rx.Observable<Y>, completed: Rx.Subject<Y>): void;
}
export declare function start<T>(argv: any): QueueObservable<T>;
export default Queue;
