"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var Rx = require("rxjs");
exports.State = {
    OK: 'OK',
    ERROR: 'ERROR'
};
var QEntryRun = /** @class */ (function () {
    function QEntryRun() {
        this.started = new Date();
    }
    QEntryRun.prototype.error = function (err) {
        this._error = err;
    };
    QEntryRun.prototype.completed = function () {
        this._completed = new Date();
    };
    return QEntryRun;
}());
exports.QEntryRun = QEntryRun;
var QEntry = /** @class */ (function () {
    function QEntry(input, output) {
        this.runs = [];
        this.input = input;
        this.output = output;
        this.created = new Date();
    }
    QEntry.prototype.isCompleted = function () {
        return !!this.runs.find(function (r) { return !!r.completed; });
    };
    QEntry.prototype.isWaiting = function () {
        // console.log(this.runs);
        return this.runs.length == 0 || !!this.runs.find(function (qer) { return !qer.completed; });
    };
    QEntry.prototype.newQEntryRun = function () {
        var qer = new QEntryRun();
        this.runs.push(qer);
        return qer;
    };
    return QEntry;
}());
exports.QEntry = QEntry;
var LogMsg = /** @class */ (function () {
    function LogMsg(level, parts) {
        this.level = level;
        this.parts = parts;
    }
    return LogMsg;
}());
exports.LogMsg = LogMsg;
var WorkerState = /** @class */ (function () {
    function WorkerState(q, qw) {
        this.q = q;
        this.qWorker = qw;
        this.running = null;
    }
    WorkerState.prototype.isFree = function () {
        return !this.running;
    };
    WorkerState.prototype.run = function (qe) {
        var _this = this;
        this.running = qe.newQEntryRun();
        this.output = new Rx.Subject();
        this.output.subscribe(function (a) {
            // console.log('WorkerState:next', a);
            qe.output.next(a);
        }, function (err) {
            // console.log('WorkerState:error', err);
            _this.running.error(err);
            _this.running = null;
            qe.output.error(err);
        }, function () {
            // console.log('WorkerState:completed');
            _this.running.completed();
            _this.running = null;
            qe.output.complete();
            _this.q.processMemoryQ();
        });
        this.qWorker(qe.input, this.output);
    };
    return WorkerState;
}());
exports.WorkerState = WorkerState;
var Logger = /** @class */ (function () {
    function Logger(upStream, qid) {
        this.upStream = upStream;
        this.base = ["Q[" + qid + "]:"];
    }
    Logger.prototype.info = function () {
        var args = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            args[_i] = arguments[_i];
        }
        this.upStream.next(new LogMsg('info', this.base.concat(args)));
    };
    Logger.prototype.error = function () {
        var args = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            args[_i] = arguments[_i];
        }
        this.upStream.next(new LogMsg('error', this.base.concat(args)));
    };
    Logger.prototype.debug = function () {
        var args = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            args[_i] = arguments[_i];
        }
        this.upStream.next(new LogMsg('debug', this.base.concat(args)));
    };
    return Logger;
}());
exports.Logger = Logger;
var Queue = /** @class */ (function () {
    function Queue(obs, argv, id) {
        // public q: Rx.Subject<QEntry<T>>;
        // public readonly deadLetter: Rx.Subject<QEntry<T>>;
        this.qEntries = [];
        // this.inProcess = false;
        // this.obs = obs;
        this.logger = new Logger(obs, id);
        this.lengthSubject = new Rx.Subject();
        this.taskTimer = argv.taskTimer || 500;
        this.reclaimTimeout = argv.reclaimTimeout || 10000;
        this.retryWaitTime = argv.retryWaitTime || 1000;
        this.maxExecuteCnt = argv.maxExecuteCnt || 10;
        this.workers = [];
        // this.q = new Rx.Subject<QEntry<T>>();
        // this.q.subscribe(this.action.bind(this));
        // this.deadLetter = new Rx.Subject<QEntry<T>>();
        this.qTask = setInterval(this.processMemoryQ.bind(this), this.taskTimer);
    }
    Queue.prototype.addWorker = function (qw) {
        this.workers.push(new WorkerState(this, qw));
        return this;
    };
    Queue.prototype.length = function () {
        return this.lengthSubject;
    };
    Queue.prototype.stop = function () {
        var _this = this;
        this.logger.info('run stop');
        // console.log('run stop');
        return Rx.Observable.create(function (observer) {
            _this.logger.info('queue stop');
            var forceStopAfter = 3;
            var action = function () {
                if (forceStopAfter <= 0 || _this.qEntries.length == 0) {
                    clearInterval(_this.qTask);
                    _this.logger.info('Q Task stopped');
                    observer.next(null);
                    observer.complete();
                }
                else {
                    --forceStopAfter;
                    _this.logger.info('waiting for Q Task to stop', _this.qEntries.length, forceStopAfter);
                    setTimeout(action, _this.taskTimer);
                }
            };
            action();
        });
    };
    // private action(qe: QEntry<T>): void {
    //   this.logger.debug('action:', qe);
    //   // qe.running = new Date();
    //   // ++qe.executeCnt;
    //   // qe.task.subscribe((_qe: T) => {
    //   //   qe.running = null;
    //   //   qe.completed = new Date();
    //   //   this.logger.debug('action:completed:', qe);
    //   // }, (error: any) => {
    //   //   qe.running = null;
    //   //   qe.retryAt = new Date((new Date()).getTime() + this.retryWaitTime);
    //   //   this.logger.debug('action:error:', qe.created.getTime(),
    //   //                                      qe.retryAt.getTime(),
    //   //                                      qe);
    //   // });
    // }
    Queue.prototype.processMemoryQ = function () {
        var freeWorker = this.workers.find(function (w) { return w.isFree(); });
        var qentry = this.qEntries.find(function (q) { return q.isWaiting(); });
        if (freeWorker && qentry) {
            // this.logger.info('processMemoryQ:run');
            freeWorker.run(qentry);
        }
        var prevLen = this.qEntries.length;
        this.qEntries = this.qEntries.filter(function (a) { return !a.isCompleted(); });
        if (prevLen != this.qEntries.length) {
            this.lengthSubject.next(this.qEntries.length);
        }
        // this.qEntries = this.qEntries.filter((qe) => {
        //   if (qe.completed) {
        //     this.logger.debug('completed remove', qe);
        //     return false;
        //   } else if (qe.executeCnt >= this.maxExecuteCnt) {
        //     this.logger.error('drop queue entry:', qe);
        //     this.deadLetter.next(qe);
        //     return false;
        //   } else if (qe.retryAt) {
        //     if (now >= qe.retryAt.getTime()) {
        //       qe.retryAt = null;
        //       this.logger.debug('retry queue entry:', qe);
        //       this.q.next(qe);
        //     }
        //     return true;
        //   } else if (!qe.running) {
        //     this.logger.debug('trigger action:', qe);
        //     this.q.next(qe);
        //   } else if ((qe.running && qe.executeCnt < this.maxExecuteCnt &&
        //     (now - qe.running.getTime()) >= this.reclaimTimeout)) {
        //     qe.running = null;
        //     this.logger.debug('retry execution time entry:', qe);
        //     this.q.next(qe);
        //   }
        //   return true;
        // });
    };
    Queue.prototype.push = function (action, completed) {
        this.qEntries.push(new QEntry(action, completed));
        this.lengthSubject.next(this.qEntries.length);
        this.processMemoryQ();
    };
    return Queue;
}());
exports.Queue = Queue;
var queueId = ~~(0x1000000 * Math.random());
function start(argv) {
    return Rx.Observable.create(function (obs) {
        var q = new Queue(obs, argv, queueId++);
        q.logger.info('Queue Started');
        obs.next(q);
    });
}
exports.start = start;
exports.default = Queue;
//# sourceMappingURL=simqle.js.map